# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains logging utilities."""

from copy import copy, deepcopy
from typing import Callable, Iterable, Optional, Union

from .algorithms.searchalgorithm import SearchAlgorithm
from .core import Problem, Solution
from .tools import is_sequence

try:
    import pandas
except ImportError:
    pandas = None


try:
    import sacred
except ImportError:
    sacred = None


try:
    import mlflow
except ImportError:
    mlflow = None


try:
    import neptune.new as neptune
except ImportError:
    neptune = None


class Logger:
    """Base class for all logging classes."""

    def __init__(self, searcher: SearchAlgorithm, *, interval: int = 1, after_first_step: bool = False):
        """`__init__(...)`: Initialize the Logger.

        Args:
            searcher: The evolutionary algorithm instance whose progress
                is to be logged.
            interval: Expected as an integer n.
                Logging is to be done at every n iterations.
            after_first_step: Expected as a boolean.
                Meaningful only if interval is set as an integer greater
                than 1. Let us suppose that interval is set as 10.
                If after_first_step is False (which is the default),
                then the logging will be done at steps 10, 20, 30, and so on.
                On the other hand, if after_first_step is True,
                then the logging will be done at steps 1, 11, 21, 31, and so
                on.
        """
        searcher.log_hook.append(self)
        self._interval = int(interval)
        self._after_first_step = bool(after_first_step)
        self._steps_count = 0

    def __call__(self, status: dict):
        if self._after_first_step:
            n = self._steps_count
            self._steps_count += 1
        else:
            self._steps_count += 1
            n = self._steps_count

        if (n % self._interval) == 0:
            self._log(self._filter(status))

    def _filter(self, status: dict) -> dict:
        return status

    def _log(self, status: dict):
        raise NotImplementedError


class ScalarLogger(Logger):
    def _filter(self, status: dict) -> dict:
        short_status = {}
        for k, v in status.items():
            if (not is_sequence(v)) and (not isinstance(v, Solution)):
                short_status[k] = v
        return short_status


class StdOutLogger(ScalarLogger):
    """A logger which prints the status into the screen."""

    def __init__(
        self,
        searcher: SearchAlgorithm,
        *,
        interval: int = 1,
        after_first_step: bool = False,
        leading_keys: Iterable[str] = ("iter",),
    ):
        """`__init__(...)`: Initialize the StdOutLogger.

        Args:
            searcher: The evolutionary algorithm instance whose progress
                is to be logged.
            interval: Expected as an integer n.
                Logging is to be done at every n iterations.
            after_first_step: Expected as a boolean.
                Meaningful only if interval is set as an integer greater
                than 1. Let us suppose that interval is set as 10.
                If after_first_step is False (which is the default),
                then the logging will be done at steps 10, 20, 30, and so on.
                On the other hand, if after_first_step is True,
                then the logging will be done at steps 1, 11, 21, 31, and so
                on.
            leading_keys: A sequence of strings where each string is a status
                key. When printing the status, these keys will be shown first.
        """
        super().__init__(searcher, interval=interval, after_first_step=after_first_step)
        self._leading_keys = list(leading_keys)
        self._leading_keys_set = set(self._leading_keys)

    def _log(self, status: dict):
        max_key_length = max([len(str(k)) for k in status.keys()])

        def report(k, v):
            nonlocal max_key_length
            print(str(k).rjust(max_key_length), ":", v)

        for k in self._leading_keys:
            if k in status:
                v = status[k]
                report(k, v)
        for k, v in status.items():
            if k not in self._leading_keys_set:
                report(k, v)
        print()


if pandas is not None:

    class PandasLogger(ScalarLogger):
        """A logger which collects status information and
        generates a pandas.DataFrame at the end.
        """

        def __init__(self, searcher: SearchAlgorithm, *, interval: int = 1, after_first_step: bool = False):
            """`__init__(...)`: Initialize the PandasLogger.

            Args:
                searcher: The evolutionary algorithm instance whose progress
                    is to be logged.
                interval: Expected as an integer n.
                    Logging is to be done at every n iterations.
                after_first_step: Expected as a boolean.
                    Meaningful only if interval is set as an integer greater
                    than 1. Let us suppose that interval is set as 10.
                    If after_first_step is False (which is the default),
                    then the logging will be done at steps 10, 20, 30, and
                    so on. On the other hand, if after_first_step is True,
                    then the logging will be done at steps 1, 11, 21, 31, and
                    so on.
            """
            super().__init__(searcher, interval=interval, after_first_step=after_first_step)
            self._data = []

        def _log(self, status: dict):
            self._data.append(deepcopy(status))

        def to_dataframe(self, *, index: Optional[str] = "iter") -> pandas.DataFrame:
            """Generate a pandas.DataFrame from the collected
            status information.

            Args:
                index: The column to be set as the index.
                    If passed as None, then no index will be set.
                    The default is "iter".
            """
            result = pandas.DataFrame(self._data)
            if index is not None:
                result.set_index(index, inplace=True)
            return result


if sacred is not None:

    ExpOrRun = Union[sacred.Experiment, sacred.run.Run]

    class SacredLogger(ScalarLogger):
        """A logger which stores the status via the Run object of sacred."""

        def __init__(
            self,
            searcher: SearchAlgorithm,
            run: ExpOrRun,
            result: Optional[str] = None,
            *,
            interval: int = 1,
            after_first_step: bool = False,
        ):
            """`__init__(...)`: Initialize the SacredLogger.

            Args:
                searcher: The evolutionary algorithm instance whose progress
                    is to be logged.
                run: An instance of `sacred.run.Run` or `sacred.Experiment`,
                    using which the progress will be logged.
                result: The key in the status dictionary whose associated
                    value will be registered as the current result
                    of the experiment.
                    If left as None, no result will be registered.
                interval: Expected as an integer n.
                    Logging is to be done at every n iterations.
                after_first_step: Expected as a boolean.
                    Meaningful only if interval is set as an integer greater
                    than 1. Let us suppose that interval is set as 10.
                    If after_first_step is False (which is the default),
                    then the logging will be done at steps 10, 20, 30, and
                    so on. On the other hand, if after_first_step is True,
                    then the logging will be done at steps 1, 11, 21, 31,
                    and so on.
            """
            super().__init__(searcher, interval=interval, after_first_step=after_first_step)
            self._result = result
            self._run = run

        def _log(self, status: dict):
            for k, v in status.items():
                self._run.log_scalar(k, v)
            if self._result is not None:
                self._run.result = status[self._result]


if mlflow is not None:
    MlflowID = Union[str, bytes, int]

    class MlflowLogger(ScalarLogger):
        """A logger which stores the status via Mlflow."""

        def __init__(
            self,
            searcher: SearchAlgorithm,
            client: Optional[mlflow.tracking.MlflowClient] = None,
            run: Union[mlflow.entities.Run, Optional[MlflowID]] = None,
            *,
            interval: int = 1,
            after_first_step: bool = False,
        ):
            """`__init__(...)`: Initialize the MlflowLogger.

            Args:
                searcher: The evolutionary algorithm instance whose progress
                    is to be logged.
                client: The MlflowClient object whose log_metric() method
                    will be used for logging. This can be passed as None,
                    in which case mlflow.log_metrics() will be used instead.
                    Please note that, if a client is provided, the `run`
                    argument is required as well.
                run: Expected only if a client is provided.
                    This is the mlflow Run object (an instance of
                    mlflow.entities.Run), or the ID of the mlflow run.
                interval: Expected as an integer n.
                    Logging is to be done at every n iterations.
                after_first_step: Expected as a boolean.
                    Meaningful only if interval is set as an integer greater
                    than 1. Let us suppose that interval is set as 10.
                    If after_first_step is False (which is the default),
                    then the logging will be done at steps 10, 20, 30, and
                    so on. On the other hand, if after_first_step is True,
                    then the logging will be done at steps 1, 11, 21, 31,
                    and so on.
            """

            super().__init__(searcher, interval=interval, after_first_step=after_first_step)

            self._client = client
            self._run_id: Optional[MlflowID] = None

            if self._client is None:
                if run is not None:
                    raise ValueError("Received `run`, but `client` is missing")
            else:
                if run is None:
                    raise ValueError("Received `client`, but `run` is missing")
                if isinstance(run, mlflow.entities.Run):
                    self._run_id = run.info.run_id
                else:
                    self._run_id = run

        def _log(self, status: dict):
            if self._client is None:
                mlflow.log_metrics(status, step=self._steps_count)
            else:
                for k, v in status.items():
                    self._client.log_metric(self._run_id, k, v, step=self._steps_count)


if neptune is not None:

    class NeptuneLogger(ScalarLogger):
        """A logger which stores the status via neptune."""

        def __init__(
            self,
            searcher: SearchAlgorithm,
            run,
            *,
            interval: int = 1,
            after_first_step: bool = False,
            group: Optional[str] = None,
        ):
            """`__init__(...)`: Initialize the NeptuneLogger.

            Args:
                searcher: The evolutionary algorithm instance whose progress
                    is to be logged.
                run: A `neptune.new.run.Run` instance using which the status
                    will be logged.
                interval: Expected as an integer n.
                    Logging is to be done at every n iterations.
                after_first_step: Expected as a boolean.
                    Meaningful only if interval is set as an integer greater
                    than 1. Let us suppose that interval is set as 10.
                    If after_first_step is False (which is the default),
                    then the logging will be done at steps 10, 20, 30, and so on.
                    On the other hand, if after_first_step is True,
                    then the logging will be done at steps 1, 11, 21, 31, and so
                    on.
                group: Into which group will the metrics be stored.
                    For example, if the status keys to be logged are "score" and
                    "elapsed", and `group` is set as "training", then the metrics
                    will be sent to neptune with the keys "training/score" and
                    "training/elapsed". `group` can also be left as None,
                    in which case the status will be sent to neptune with the
                    key names unchanged.
            """
            super().__init__(searcher, interval=interval, after_first_step=after_first_step)
            self._run = run
            self._group = group

        def _log(self, status: dict):
            for k, v in status.items():
                target_key = k if self._group is None else self._group + "/" + k
                self._run[target_key].log(v)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import torch

import evotorch as et


class MyProblem(et.Problem):
    def __init__(self, num_actors: int):
        super().__init__(
            objective_sense="min",
            initial_bounds=(-10.0, 10.0),
            solution_length=5,
            num_actors=int(num_actors),
        )

    def _evaluate_batch(self, batch: et.SolutionBatch):
        x = batch.values
        norms = torch.linalg.norm(x, dim=-1)
        batch.set_evals(norms)

    def get_actor_index(self) -> int:
        return self.actor_index

    def get_env(self):
        class DummyEnv:
            def __init__(self, actor_index: int):
                self.actor_index = actor_index

            def get_parent_actor_index(self) -> int:
                return self.actor_index

        result = DummyEnv(self.actor_index)
        return result


def test_remote_access():
    num_actors = 4
    problem = MyProblem(num_actors=num_actors)
    expected_indices = set(range(num_actors))
    all_actor_indices = set(problem.all_remote_problems().get_actor_index())
    all_env_actor_indices = set(problem.all_remote_envs().get_parent_actor_index())
    assert all_actor_indices == expected_indices
    assert all_env_actor_indices == expected_indices
    assert len(problem.actors) == num_actors



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This namespace contains the `NeuroevolutionProblem` class."""

import math
from copy import deepcopy
from typing import Any, Callable, Iterable, List, Optional, Union

import ray
import torch
from torch import nn

from ..core import BoundsPairLike, DType, ObjectiveSense, Problem, Solution
from ..tools.misc import Device, is_sequence
from .net.misc import count_parameters, fill_parameters
from .net.parser import str_to_net


class NEProblem(Problem):
    """
    Base class for neuro-evolution problems where the goal is to optimize the
    parameters of a neural network represented as a PyTorch module.

    Any problem inheriting from this class is expected to override the method
    `_evaluate_network(self, net: torch.nn.Module) -> Union[torch.Tensor, float]`
    where `net` is the neural network to be evaluated, and the return value
    is a scalar or a vector (for multi-objective cases) expressing the
    fitness value(s).

    Alternatively, this class can be directly instantiated in the following
    way:

    ```python
    def f(module: MyTorchModuleClass) -> Union[float, torch.Tensor, tuple]:
        # Evaluate the given PyTorch module here
        fitness = ...
        return fitness

    problem = NEProblem(
        "min", MyTorchModuleClass, f,
        ...
    )
    ```

    which specifies that the problem's goal is to minimize the return of the
    function `f`.
    For multi-objective cases, the fitness returned by `f` is expected as a
    1-dimensional tensor. For when the problem has additional evaluation data,
    a two-element tuple can be returned by `f` instead, where the first
    element is the fitness value(s) and the second element is a 1-dimensional
    tensor storing the additional data.
    """

    def __init__(
        self,
        objective_sense: ObjectiveSense,
        network: Union[str, nn.Module, Callable[[], nn.Module]],
        network_eval_func: Optional[Callable] = None,
        *,
        network_args: Optional[dict] = None,
        initial_bounds: Optional[BoundsPairLike] = (-0.00001, 0.00001),
        eval_dtype: Optional[DType] = None,
        eval_data_length: int = 0,
        seed: Optional[int] = None,
        num_actors: Optional[Union[int, str]] = "num_devices",
        actor_config: Optional[dict] = None,
        num_gpus_per_actor: Optional[Union[int, float, str]] = None,
        num_subbatches: Optional[int] = None,
        subbatch_size: Optional[int] = None,
        device: Optional[Device] = None,
    ):
        """
        `__init__(...)`: Initialize the NEProblem.

        Args:
            objective_sense: The objective sense, expected as "min" or "max"
                for single-objective cases, or as a sequence of strings
                (each string being "min" or "max") for multi-objective cases.
            network: A network structure string, or a Callable (which can be
                a class inheriting from `torch.nn.Module`, or a function
                which returns a `torch.nn.Module` instance), or an instance
                of `torch.nn.Module`.
                The object provided here determines the structure of the
                neural network whose parameters will be evolved.
                A network structure string is a string which can be processed
                by `evotorch.neuroevolution.net.str_to_net(...)`.
                Please see the documentation of the function
                `evotorch.neuroevolution.net.str_to_net(...)` to see how such
                a neural network structure string looks like.
            network_eval_func: Optionally a function (or any Callable object)
                which receives a PyTorch module as its argument, and returns
                either a fitness, or a two-element tuple containing the fitness
                and the additional evaluation data. The fitness can be a scalar
                (for single-objective cases) or a 1-dimensional tensor (for
                multi-objective cases). The additional evaluation data is
                expected as a 1-dimensional tensor.
                If this argument is left as None, it will be expected that
                the method `_evaluate_network(...)` is overriden by the
                inheriting class.
            network_args: Optionally a dict-like object, storing keyword
                arguments to be passed to the network while instantiating it.
            initial_bounds: Specifies an interval from which the values of the
                initial neural network parameters will be drawn.
            eval_dtype: dtype to be used for fitnesses. If not specified, then
                `eval_dtype` will be inferred from the dtype of the parameters
                of the neural network.
                In more details, if the neural network's parameters have a
                float dtype, `eval_dtype` will be a compatible float.
                Otherwise, it will be "float32".
            eval_data_length: Length of the extra evaluation data.
            seed: Random number seed. If left as None, this NEProblem instance
                will not have its own random generator, and the global random
                generator of PyTorch will be used instead.
            num_actors: Number of actors to create for parallelized
                evaluation of the solutions.
                Certain string values are also accepted.
                When given as "max" or as "num_cpus", the number of actors
                will be equal to the number of all available CPUs in the ray
                cluster.
                When given as "num_gpus", the number of actors will be
                equal to the number of all available GPUs in the ray
                cluster, and each actor will be assigned a GPU.
                When given as "num_devices", the number of actors will be
                equal to the minimum among the number of CPUs and the number
                of GPUs available in the cluster (or will be equal to the
                number of CPUs if there is no GPU), and each actor will be
                assigned a GPU (if available).
                If `num_actors` is given as "num_gpus" or "num_devices",
                the argument `num_gpus_per_actor` must not be used,
                and the `actor_config` dictionary must not contain the
                key "num_gpus".
                If `num_actors` is given as something other than "num_gpus"
                or "num_devices", and if you wish to assign GPUs to each
                actor, then please see the argument `num_gpus_per_actor`.
            actor_config: A dictionary, representing the keyword arguments
                to be passed to the options(...) used when creating the
                ray actor objects. To be used for explicitly allocating
                resources per each actor.
                For example, for declaring that each actor is to use a GPU,
                one can pass `actor_config=dict(num_gpus=1)`.
                Can also be given as None (which is the default),
                if no such options are to be passed.
            num_gpus_per_actor: Number of GPUs to be allocated by each
                remote actor.
                The default behavior is to NOT allocate any GPU at all
                (which is the default behavior of the ray library as well).
                When given as a number `n`, each actor will be given
                `n` GPUs (where `n` can be an integer, or can be a `float`
                for fractional allocation).
                When given as a string "max", then the available GPUs
                across the entire ray cluster (or within the local computer
                in the simplest cases) will be equally distributed among
                the actors.
                When given as a string "all", then each actor will have
                access to all the GPUs (this will be achieved by suppressing
                the environment variable `CUDA_VISIBLE_DEVICES` for each
                actor).
                When the problem is not distributed (i.e. when there are
                no actors), this argument is expected to be left as None.
            num_subbatches: If `num_subbatches` is None (assuming that
                `subbatch_size` is also None), then, when evaluating a
                population, the population will be split into n pieces, `n`
                being the number of actors, and each actor will evaluate
                its assigned piece. If `num_subbatches` is an integer `m`,
                then the population will be split into `m` pieces,
                and actors will continually accept the next unevaluated
                piece as they finish their current tasks.
                The arguments `num_subbatches` and `subbatch_size` cannot
                be given values other than None at the same time.
                While using a distributed algorithm, this argument determines
                how many sub-batches will be generated, and therefore,
                how many gradients will be computed by the remote actors.
            subbatch_size: If `subbatch_size` is None (assuming that
                `num_subbatches` is also None), then, when evaluating a
                population, the population will be split into `n` pieces, `n`
                being the number of actors, and each actor will evaluate its
                assigned piece. If `subbatch_size` is an integer `m`,
                then the population will be split into pieces of size `m`,
                and actors will continually accept the next unevaluated
                piece as they finish their current tasks.
                When there can be significant difference across the solutions
                in terms of computational requirements, specifying a
                `subbatch_size` can be beneficial, because, while one
                actor is busy with a subbatch containing computationally
                challenging solutions, other actors can accept more
                tasks and save time.
                The arguments `num_subbatches` and `subbatch_size` cannot
                be given values other than None at the same time.
                While using a distributed algorithm, this argument determines
                the size of a sub-batch (or sub-population) sampled by a
                remote actor for computing a gradient.
                In distributed mode, it is expected that the population size
                is divisible by `subbatch_size`.
            device: Default device in which a new population will be generated
                and the neural networks will operate.
                If not specified, "cpu" will be used.
        """
        # Set the main device of the problem
        # Although the operation of setting the main device is done by the main Problem class,
        # here we need this at an earlier stage.
        if device is None:
            device = "cpu"
        self._device = torch.device(device)

        # Set the network
        self._original_network = network
        self._network_args = {} if network_args is None else deepcopy(network_args)
        if isinstance(self._original_network, nn.Module):
            self._original_network = self._original_network.cpu()

        # Store the function that will evaluate the network, if available
        self._network_eval_func: Optional[Callable] = network_eval_func

        self.instantiated_network: nn.Module = None

        # Create temporary network
        temp_network = self._instantiate_net(self._original_network, device="cpu")

        super().__init__(
            objective_sense=objective_sense,
            initial_bounds=initial_bounds,
            bounds=None,  # Neuroevolution is an unbounded problem
            solution_length=count_parameters(temp_network),  # The solution length is inherited from the network passed
            dtype=next(temp_network.parameters()).dtype,  # The datatype is inherited from the network passed
            eval_dtype=eval_dtype,
            device=device,
            eval_data_length=eval_data_length,
            seed=seed,
            num_actors=num_actors,
            num_gpus_per_actor=num_gpus_per_actor,
            actor_config=actor_config,
            num_subbatches=num_subbatches,
            subbatch_size=subbatch_size,
            store_solution_stats=None,
        )

    @property
    def network_device(self) -> Device:
        """The device on which the problem should place data e.g. the network"""
        cpu_device = torch.device("cpu")
        if self.is_main:
            # This is the case where this is the main process (not a remote actor)
            if self.device == cpu_device:
                # If the main device of the problem is "cpu", then we assume that the network is going to be on the cpu as well
                return cpu_device
            else:
                # If the main device of the problem is some other device, then it is that device into which the network will be put
                return self.device
        else:
            # If this is a remote actor, then the network will be put into the auxiliary device allocated for that actor
            return self.aux_device

    @property
    def _network_constants(self) -> dict:
        """Named constants which can be passed to the network instantiation e.g. input/output dimension. To be overridden by the user for custom fixed constants for a problem"""
        return {}

    def network_constants(self) -> dict:
        """Named constants which can be passed to the network instantiation e.g. input/output dimension"""
        constants = {}
        constants.update(self._network_constants)
        constants.update(self._network_args)
        return constants

    @property
    def _nonserialized_attribs(self) -> List[str]:
        return ["instantiated_network"]

    def _instantiate_net(self, network: Union[str, nn.Module, dict], device: Optional[Device] = None) -> nn.Module:
        """Instantiate the network on the target device, to be overridden by the user for custom behaviour
        Returns:
            instantiated_network (nn.Module): The network instantiated on the target device
        """
        # Branching point determines instantiation of network
        if isinstance(network, str):
            # Passed argument was a string representation of a torch module
            instantiated_network = str_to_net(network, **self.network_constants())
        elif isinstance(network, nn.Module):
            # Passed argument was directly a torch module
            instantiated_network = network
        else:
            # Passed argument was callable yielding network
            instantiated_network = network(**self.network_constants())

        # Map to device
        device = self.network_device if device is None else device
        instantiated_network = instantiated_network.to(device)

        return instantiated_network

    def _prepare(self) -> None:
        """Instantiate the network on the target device, if not already done"""
        self.instantiated_network = self._instantiate_net(self._original_network)
        # Clear reference to original network
        self._original_network = None

    def make_net(self, parameters: Iterable) -> nn.Module:
        """
        Make a new network filled with the provided parameters.

        Args:
            parameters: Parameters to be used as weights within the network.
                Can be a Solution, or any 1-dimensional Iterable that can be
                converted to a PyTorch tensor.
        Returns:
            A new network, as a `torch.Module` instance.
        """
        if isinstance(parameters, Solution):
            parameters = parameters.access_values(keep_evals=True)
        else:
            parameters = self.as_tensor(parameters)
        with torch.no_grad():
            net = deepcopy(self.parameterize_net(parameters))
        return net

    def parameterize_net(self, parameters: torch.Tensor) -> nn.Module:
        """Parameterize the network with a given set of parameters.
        Args:
            parameters (torch.Tensor): The parameters with which to instantiate the network
        Returns:
            instantiated_network (nn.Module): The network instantiated with the parameters
        """
        # Check if network exists
        if self.instantiated_network is None:
            self.instantiated_network = self._instantiate_net(self._original_network)

        network = self.instantiated_network

        # Move the parameters if needed
        if parameters.device != self.network_device:
            parameters = parameters.to(self.network_device)

        # Fill the network with the parameters
        fill_parameters(network, parameters)

        # Return the network
        return network

    @property
    def _grad_device(self) -> Device:
        """
        Get the device in which new solutions will be made in distributed mode.

        In more details, in distributed mode, each actor creates its own
        sub-populations, evaluates them, and computes its own gradient
        (all such actor gradients eventually being collected by the
        distribution-based search algorithm in the main process).
        For some problem types, it can make sense for the remote actors to
        create their temporary sub-populations on another device
        (e.g. on the GPU that is allocated specifically for them).
        For such situations, one is encouraged to override this property
        and make it return whatever device is to be used.

        In the case of NEProblem, this property returns whatever device
        is specified by the property `network_device`.
        """
        return self.network_device

    def _evaluate_network(self, network: nn.Module) -> Union[float, torch.Tensor, tuple]:
        """
        Evaluate a network and return the evaluation result(s).

        In the case where the `__init__` of `NEProblem` was not given
        a network evaluator function (via the argument `network_eval_func`),
        it will be expected that the inheriting class overrides this
        method and defines how a network should be evaluated.

        Args:
            network (nn.Module): The network to evaluate
        Returns:
            fitness: The networks' fitness value(s), as a scalar for
                single-objective cases, or as a 1-dimensional tensor
                for multi-objective cases. The returned value can also
                be a two-element tuple where the first element is the
                fitness (as a scalar or as a vector) and the second
                element is a 1-dimensional vector storing the extra
                evaluation data.
        """
        raise NotImplementedError

    def _evaluate(self, solution: Solution):
        """
        Evaluate a single solution.
        This is achieved by parameterising the problem's attribute
        named `instantiated_network`, and then evaluating the network
        with the method `_evaluate_network(...)`.

        Args:
            solution (Solution): The solution to evaluate.
        """
        parameters = solution.values

        if self._network_eval_func is None:
            evaluator = self._evaluate_network
        else:
            evaluator = self._network_eval_func

        fitnesses = evaluator(self.parameterize_net(parameters))

        if isinstance(fitnesses, tuple):
            solution.set_evals(*fitnesses)
        else:
            solution.set_evals(fitnesses)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains operators defined to work with problems
whose `dtype`s are real numbers (e.g. `torch.float32`).
"""

from copy import deepcopy
from typing import Iterable, Optional, Union

import torch

from ..core import Problem, SolutionBatch
from ..tools.ranking import rank
from .base import CopyingOperator, CrossOver, SingleObjOperator


class GaussianMutation(CopyingOperator):
    """
    Gaussian mutation operator.

    Follows the algorithm description in:

        Sean Luke, 2013, Essentials of Metaheuristics, Lulu, second edition
        available for free at http://cs.gmu.edu/~sean/book/metaheuristics/
    """

    def __init__(self, problem: Problem, *, stdev: float, mutation_probability: Optional[float] = 1.0):
        """
        `__init__(...)`: Initialize the GaussianMutation.

        Args:
            problem: The problem object to work with.
            stdev: The standard deviation of the Gaussian noise to apply on
                each decision variable.
            mutation_probability: The probability of mutation, for each
                decision variable.
                By default, the value of this argument is 1.0, which means
                that all of the decision variables will be affected by the
                mutation.
        """

        super().__init__(problem)
        self._mutation_probability = float(mutation_probability)
        self._stdev = float(stdev)

    @torch.no_grad()
    def _do(self, batch: SolutionBatch) -> SolutionBatch:
        result = deepcopy(batch)
        data = result.access_values()
        mutation_matrix = self.problem.make_uniform_shaped_like(data) <= self._mutation_probability
        data[mutation_matrix] += self._stdev * self.problem.make_gaussian_shaped_like(data[mutation_matrix])
        data[:] = self._respect_bounds(data)
        return result


class OnePointCrossOver(CrossOver):
    """
    Representation of a one-point cross-over operator.

    When this operator is applied on a SolutionBatch,
    a tournament selection technique is used for selecting
    parent solutions from the batch, and then those parent
    solutions are mated via cutting from a random position
    and recombining. The result of these recombination
    operations is a new SolutionBatch, containing the children
    solutions. The original SolutionBatch stays unmodified.
    """

    def __init__(
        self,
        problem: Problem,
        *,
        tournament_size: int,
        obj_index: Optional[int] = None,
        num_children: Optional[int] = None,
        cross_over_rate: Optional[float] = None,
    ):
        """
        `__init__(...)`: Initialize the OnePointCrossOver.

        Args:
            problem: The problem object to work on.
            tournament_size: What is the size (or length) of a tournament
                when selecting a parent candidate from a population
            obj_index: Objective index according to which the selection
                will be done.
            num_children: Optionally a number of children to produce by the
                cross-over operation.
                Not to be used together with `cross_over_rate`.
                If `num_children` and `cross_over_rate` are both None,
                then the number of children is equal to the number
                of solutions received.
            cross_over_rate: Optionally expected as a real number between
                0.0 and 1.0. Specifies the number of cross-over operations
                to perform. 1.0 means `1.0 * len(solution_batch)` amount of
                cross overs will be performed, resulting in
                `2.0 * len(solution_batch)` amount of children.
                Not to be used together with `num_children`.
                If `num_children` and `cross_over_rate` are both None,
                then the number of children is equal to the number
                of solutions received.
        """

        super().__init__(
            problem,
            tournament_size=tournament_size,
            obj_index=obj_index,
            num_children=num_children,
            cross_over_rate=cross_over_rate,
        )

    @torch.no_grad()
    def _do_cross_over(self, parents1: torch.Tensor, parents2: torch.Tensor) -> SolutionBatch:
        # What we expect here is this:
        #
        #    parents1      parents2
        #    ==========    ==========
        #    parents1[0]   parents2[0]
        #    parents1[1]   parents2[1]
        #    ...           ...
        #    parents1[N]   parents2[N]
        #
        # where parents1 and parents2 are 2D tensors, each containing values of N solutions.
        # For each row i, we will apply cross-over on parents1[i] and parents2[i].
        # From each cross-over, we will obtain 2 children.
        # This means, there are N pairings, and 2N children.

        num_pairings = parents1.shape[0]
        # num_children = num_pairings * 2

        device = parents1[0].device
        dtype = parents1[0].dtype
        solution_length = len(parents1[0])

        # For each pairing, generate a gene index at which the parent solutions will be cut and recombined
        crossover_point = self.problem.make_randint((num_pairings, 1), n=(solution_length - 1), device=device) + 1

        # For each pairing, generate all gene indices (i.e. [0, 1, 2, ...] for each pairing)
        gene_indices = (
            torch.arange(0, solution_length, device=device).unsqueeze(0).expand(num_pairings, solution_length)
        )

        # Make a mask for crossing over. (0: take the value from one parent, 1: take the value from the other parent)
        # For gene indices less than crossover_point of that pairing, the mask takes the value 0.
        # Otherwise, the mask takes the value 1.
        crossover_mask = (gene_indices >= crossover_point).to(dtype)

        # Using the mask, generate two children.
        children1 = crossover_mask * parents1 + (1 - crossover_mask) * parents2
        children2 = crossover_mask * parents2 + (1 - crossover_mask) * parents1

        # Combine the children tensors in one big tensor
        children = torch.cat([children1, children2], dim=0)

        # Write the children solutions into a new SolutionBatch, and return the new batch
        result = self._make_children_batch(children)
        return result


class SimulatedBinaryCrossOver(CrossOver):
    """
    Representation of a simulated binary cross-over (SBX).

    When this operator is applied on a SolutionBatch,
    a tournament selection technique is used for selecting
    parent solutions from the batch, and then those parent
    solutions are mated via SBX. The generated children
    solutions are given in a new SolutionBatch.
    The original SolutionBatch stays unmodified.

    Reference:

        Kalyanmoy Deb, Hans-Georg Beyer (2001).
        Self-Adaptive Genetic Algorithms with Simulated Binary Crossover.
    """

    def __init__(
        self,
        problem: Problem,
        *,
        tournament_size: int,
        eta: float,
        obj_index: Optional[int] = None,
        num_children: Optional[int] = None,
        cross_over_rate: Optional[float] = None,
    ):
        """
        `__init__(...)`: Initialize the SimulatedBinaryCrossOver.

        Args:
            problem: Problem object to work with.
            tournament_size: What is the size (or length) of a tournament
                when selecting a parent candidate from a population.
            eta: The crowding index, expected as a float.
                Bigger eta values result in children closer
                to their parents.
            obj_index: Objective index according to which the selection
                will be done.
            num_children: Optionally a number of children to produce by the
                cross-over operation.
                Not to be used together with `cross_over_rate`.
                If `num_children` and `cross_over_rate` are both None,
                then the number of children is equal to the number
                of solutions received.
            cross_over_rate: Optionally expected as a real number between
                0.0 and 1.0. Specifies the number of cross-over operations
                to perform. 1.0 means `1.0 * len(solution_batch)` amount of
                cross overs will be performed, resulting in
                `2.0 * len(solution_batch)` amount of children.
                Not to be used together with `num_children`.
                If `num_children` and `cross_over_rate` are both None,
                then the number of children is equal to the number
                of solutions received.
        """

        super().__init__(
            problem,
            tournament_size=int(tournament_size),
            obj_index=obj_index,
            num_children=num_children,
            cross_over_rate=cross_over_rate,
        )
        self._eta = float(eta)

    def _do_cross_over(self, parents1: torch.Tensor, parents2: torch.Tensor) -> SolutionBatch:
        # Generate u_i values which determine the spread
        u = self.problem.make_uniform_shaped_like(parents1)

        # Compute beta_i values from u_i values as the actual spread per dimension
        betas = (2 * u).pow(1.0 / (self._eta + 1.0))  # Compute all values for u_i < 0.5 first
        betas[u > 0.5] = (1.0 / (2 * (1.0 - u[u > 0.5]))).pow(
            1.0 / (self._eta + 1.0)
        )  # Replace the values for u_i >= 0.5
        children1 = 0.5 * (
            (1 + betas) * parents1 + (1 - betas) * parents2
        )  # Create the first set of children from the beta values
        children2 = 0.5 * (
            (1 + betas) * parents2 + (1 - betas) * parents1
        )  # Create the second set of children as a mirror of the first set of children

        # Combine the children tensors in one big tensor
        children = torch.cat([children1, children2], dim=0)

        # Respect the lower and upper bounds defined by the problem object
        children = self._respect_bounds(children)

        # Write the children solutions into a new SolutionBatch, and return the new batch
        result = self._make_children_batch(children)

        return result


class CosynePermutation(CopyingOperator):
    """
    Representation of permutation operation on a SolutionBatch.

    For each decision variable index, a permutation operation across
    all or a subset of solutions, is performed.
    The result is returned on a new SolutionBatch.
    The original SolutionBatch remains unmodified.

    Reference:

        F.Gomez, J.Schmidhuber, R.Miikkulainen (2008).
        Accelerated Neural Evolution through Cooperatively Coevolved Synapses
        Journal of Machine Learning Research 9, 937-965
    """

    def __init__(self, problem: Problem, obj_index: Optional[int] = None, *, permute_all: bool = False):
        """
        `__init__(...)`: Initialize the CosynePermutation.

        Args:
            problem: The problem object to work on.
            obj_index: The index of the objective according to which the
                candidates for permutation will be selected.
                Can be left as None if the problem is single-objective,
                or if `permute_all` is given as True (in which case there
                will be no candidate selection as the entire population will
                be subject to permutation).
            permute_all: Whether or not to apply permutation on the entire
                population, instead of using a selective permutation.
        """

        if permute_all:
            if obj_index is not None:
                raise ValueError(
                    "When `permute_all` is given as True (which seems to be the case)"
                    " `obj_index` is expected as None,"
                    " because the operator is independent of any objective and any fitness in this mode."
                    " However, `permute_all` was found to be something other than None."
                )
            self._obj_index = None
        else:
            self._obj_index = problem.normalize_obj_index(obj_index)

        super().__init__(problem)

        self._permute_all = bool(permute_all)

    @property
    def obj_index(self) -> Optional[int]:
        """Objective index according to which the operator will run.
        If `permute_all` was given as True, objectives are irrelevant, in which case
        `obj_index` is returned as None.
        If `permute_all` was given as False, the relevant `obj_index` is provided
        as an integer.
        """
        return self._obj_index

    @torch.no_grad()
    def _do(self, batch: SolutionBatch) -> SolutionBatch:
        indata = batch._data

        if not self._permute_all:
            n = batch.solution_length
            ranks = batch.utility(self._obj_index, ranking_method="centered")
            # fitnesses = batch.evals[:, self._obj_index].clone().reshape(-1)
            # ranks = rank(
            #    fitnesses, ranking_method="centered", higher_is_better=(self.problem.senses[self.obj_index] == "max")
            # )
            prob_permute = (1 - (ranks + 0.5).pow(1 / float(n))).unsqueeze(1).expand(len(batch), batch.solution_length)
        else:
            prob_permute = torch.ones_like(indata)

        perm_mask = self.problem.make_uniform_shaped_like(prob_permute) <= prob_permute

        perm_mask_sorted = torch.sort(perm_mask.to(torch.long), descending=True, dim=0)[0].to(
            torch.bool
        )  # Sort permutations

        perm_rand = self.problem.make_uniform_shaped_like(prob_permute)
        perm_rand[torch.logical_not(perm_mask)] = 1.0
        permutations = torch.argsort(perm_rand, dim=0)  # Generate permutations

        perm_sort = (
            torch.arange(0, perm_mask.shape[0], device=indata.device).unsqueeze(-1).repeat(1, perm_mask.shape[1])
        )
        perm_sort[torch.logical_not(perm_mask)] += perm_mask.shape[0] + 1
        perm_sort = torch.sort(perm_sort, dim=0)[0]  # Generate the origin of permutations

        _, permutation_columns = torch.nonzero(perm_mask_sorted, as_tuple=True)
        permutation_origin_indices = perm_sort[perm_mask_sorted]
        permutation_target_indices = permutations[perm_mask_sorted]

        newbatch = SolutionBatch(like=batch, empty=True)
        newdata = newbatch._data
        newdata[:] = indata[:]
        newdata[permutation_origin_indices, permutation_columns] = newdata[
            permutation_target_indices, permutation_columns
        ]

        return newbatch



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gym
import pybullet_envs
from gym.envs.registration import EnvSpec, register


class TinyTrajHumanoidBulletEnv(gym.Env):
    ID = "TinyTrajHumanoidBulletEnv-v0"
    ENTRY_POINT = __name__ + ":TinyTrajHumanoidBulletEnv"
    MAX_EPISODE_STEPS = 200

    def __init__(self, trajectory_length=200):
        gym.Env.__init__(self)

        self.__tlimit = trajectory_length
        self.__done = True
        self.__t = 0
        self.__contained_env = gym.make("HumanoidBulletEnv-v0", new_step_api=False)

        self.observation_space = self.__contained_env.observation_space
        self.action_space = self.__contained_env.action_space
        self.reward_range = (float("-inf"), float("inf"))

    def step(self, action):
        assert not self.__done, "Trying to progress in a finished trajectory"

        observation, _, done, info = self.__contained_env.step(action)
        self.__t += 1

        if self.__t >= self.__tlimit:
            done = True

        self.__done = done

        reward = sum(self.__contained_env.rewards[1:])
        return observation, reward, done, info

    def reset(self, **kwargs):
        self.__done = False
        self.__t = 0
        return self.__contained_env.reset(**kwargs)

    def render(self, mode="human", **kwargs):
        self.__contained_env.render(mode=mode, **kwargs)

    def close(self):
        return self.__contained_env.close()

    def seed(self, seed=None):
        return self.__contained_env.seed(seed)

    @property
    def rewards(self):
        if hasattr(self.__contained_env, "rewards"):
            return self.__contained_env.rewards
        else:
            return None

    @property
    def reward(self):
        if hasattr(self.__contained_env, "reward"):
            return self.__contained_env.reward
        else:
            return None

    @property
    def body_xyz(self):
        if hasattr(self.__contained_env.robot, "body_xyz"):
            return self.__contained_env.robot.body_xyz
        else:
            return None

    @property
    def body_rpy(self):
        if hasattr(self.__contained_env.robot, "body_rpy"):
            return self.__contained_env.robot.body_rpy
        else:
            return None

    def camera_adjust(self):
        return self.__contained_env.camera_adjust()

    @property
    def robot(self):
        return self.__contained_env


register(
    id=TinyTrajHumanoidBulletEnv.ID,
    entry_point=TinyTrajHumanoidBulletEnv.ENTRY_POINT,
    max_episode_steps=TinyTrajHumanoidBulletEnv.MAX_EPISODE_STEPS,
)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any

import numpy as np
import pytest
import torch
from torch import FloatTensor

from evotorch.testing import assert_allclose, assert_dtype_matches
from evotorch.tools import misc
from evotorch.tools.objectarray import ObjectArray


def test_expect_none():
    with pytest.raises(ValueError):
        misc.expect_none("Testing", a=1, b=2)

    with pytest.raises(ValueError):
        misc.expect_none("Testing", a=None, b=2)

    with pytest.raises(ValueError):
        misc.expect_none("Testing", a=1, b=None)

    misc.expect_none("Testing", a=None, b=None)


@pytest.mark.parametrize(
    "type_,expected,err",
    [
        (object, None, TypeError),
        (Any, None, TypeError),
        ("object", None, TypeError),
        (np.float16, torch.float16, None),
        (np.float32, torch.float32, None),
        (np.float64, torch.float64, None),
        ("float16", torch.float16, None),
        ("float32", torch.float32, None),
        ("float64", torch.float64, None),
        (np.int16, torch.int16, None),
        (np.int32, torch.int32, None),
        (np.int64, torch.int64, None),
        ("int16", torch.int16, None),
        ("int32", torch.int32, None),
        ("int64", torch.int64, None),
    ],
)
def test_to_torch_dtype(type_, expected, err):
    if err is not None:
        with pytest.raises(err):
            misc.to_torch_dtype(type_)
    else:
        assert misc.to_torch_dtype(type_) is expected


@pytest.mark.parametrize(
    "type_,expected",
    [
        (torch.int16, np.dtype("int16")),
        (torch.int32, np.dtype("int32")),
        (torch.int64, np.dtype("int64")),
        (torch.float16, np.dtype("float16")),
        (torch.float32, np.dtype("float32")),
        (torch.float64, np.dtype("float64")),
        (np.int16, np.dtype("int16")),
        (np.int32, np.dtype("int32")),
        (np.int64, np.dtype("int64")),
        (np.float16, np.dtype("float16")),
        (np.float32, np.dtype("float32")),
        (np.float64, np.dtype("float64")),
        (np.dtype("int16"), np.dtype("int16")),
        (np.dtype("int32"), np.dtype("int32")),
        (np.dtype("int64"), np.dtype("int64")),
        (np.dtype("float16"), np.dtype("float16")),
        (np.dtype("float32"), np.dtype("float32")),
        (np.dtype("float64"), np.dtype("float64")),
        ("int16", np.dtype("int16")),
        ("int32", np.dtype("int32")),
        ("int64", np.dtype("int64")),
        ("float16", np.dtype("float16")),
        ("float32", np.dtype("float32")),
        ("float64", np.dtype("float64")),
        (object, np.dtype("O")),
        ("object", np.dtype("O")),
        (np.dtype("object"), np.dtype("O")),
        (np.dtype(object), np.dtype("O")),
        (np.dtype("O"), np.dtype("O")),
        (Any, np.dtype("O")),
    ],
)
def test_to_numpy_dtype(type_, expected):
    assert misc.to_numpy_dtype(type_) == expected


@pytest.mark.parametrize(
    "type_,expected",
    [
        (Any, True),
        (object, True),
        ("Any", True),
        ("object", True),
        ("O", True),
        ("int8", False),
        (np.int8, False),
        (np.dtype("int8"), False),
        (torch.int8, False),
        ("int16", False),
        (np.int16, False),
        (np.dtype("int16"), False),
        (torch.int16, False),
        ("int32", False),
        (np.int32, False),
        (np.dtype("int32"), False),
        (torch.int32, False),
        ("int64", False),
        (np.int64, False),
        (np.dtype("int64"), False),
        (torch.int64, False),
        ("float16", False),
        (np.float16, False),
        (np.dtype("float16"), False),
        (torch.float16, False),
        ("float32", False),
        (np.float32, False),
        (np.dtype("float32"), False),
        (torch.float32, False),
        ("float64", False),
        (np.float64, False),
        (np.dtype("float64"), False),
        (torch.float64, False),
    ],
)
def test_is_dtype_object(type_, expected):
    assert misc.is_dtype_object(type_) is expected


@pytest.mark.parametrize(
    "obj,expected",
    [
        ("string", False),
        (b"bytes", False),
        (10, False),
        (10.0, False),
        ([1, 2], True),
        ((1, 2), True),
        (np.int32(0), False),
        (np.float32(0.0), False),
        (np.array(10), False),
        (np.array(10.0), False),
        (torch.tensor(10), False),
        (torch.tensor(10.0), False),
        (np.array([1, 2]), True),
        (np.array([1.0, 2.0]), True),
        (torch.tensor([1, 2]), True),
        (torch.tensor([1.0, 2.0]), True),
        (np.array([[1, 2], [3, 4]]), True),
        (np.array([[1.0, 2.0], [3.0, 4.0]]), True),
        (torch.tensor([[1, 2], [3, 4]]), True),
        (torch.tensor([[1.0, 2.0], [3.0, 4.0]]), True),
    ],
)
def test_is_sequence(obj, expected):
    assert misc.is_sequence(obj) is expected


@pytest.mark.parametrize(
    "type_,expected",
    [
        (Any, False),
        (object, False),
        ("Any", False),
        ("object", False),
        ("O", False),
        ("int8", True),
        (np.int8, True),
        (np.dtype("int8"), True),
        (torch.int8, True),
        ("int16", True),
        (np.int16, True),
        (np.dtype("int16"), True),
        (torch.int16, True),
        ("int32", True),
        (np.int32, True),
        (np.dtype("int32"), True),
        (torch.int32, True),
        ("int64", True),
        (np.int64, True),
        (np.dtype("int64"), True),
        (torch.int64, True),
        ("uint8", True),
        (np.uint8, True),
        (np.dtype("uint8"), True),
        (torch.uint8, True),
        ("uint16", True),
        (np.uint16, True),
        (np.dtype("uint16"), True),
        ("uint32", True),
        (np.uint32, True),
        (np.dtype("uint32"), True),
        ("uint64", True),
        (np.uint64, True),
        (np.dtype("uint64"), True),
        ("float16", False),
        (np.float16, False),
        (np.dtype("float16"), False),
        (torch.float16, False),
        ("float32", False),
        (np.float32, False),
        (np.dtype("float32"), False),
        (torch.float32, False),
        ("float64", False),
        (np.float64, False),
        (np.dtype("float64"), False),
        (torch.float64, False),
        ("bool", False),
        (bool, False),
        (np.bool_, False),
        (torch.bool, False),
    ],
)
def test_is_dtype_integer(type_, expected):
    assert misc.is_dtype_integer(type_) is expected


@pytest.mark.parametrize(
    "type_,expected",
    [
        (Any, False),
        (object, False),
        ("Any", False),
        ("object", False),
        ("O", False),
        ("int8", False),
        (np.int8, False),
        (np.dtype("int8"), False),
        (torch.int8, False),
        ("int16", False),
        (np.int16, False),
        (np.dtype("int16"), False),
        (torch.int16, False),
        ("int32", False),
        (np.int32, False),
        (np.dtype("int32"), False),
        (torch.int32, False),
        ("int64", False),
        (np.int64, False),
        (np.dtype("int64"), False),
        (torch.int64, False),
        ("uint8", False),
        (np.uint8, False),
        (np.dtype("uint8"), False),
        (torch.uint8, False),
        ("uint16", False),
        (np.uint16, False),
        (np.dtype("uint16"), False),
        ("uint32", False),
        (np.uint32, False),
        (np.dtype("uint32"), False),
        ("uint64", False),
        (np.uint64, False),
        (np.dtype("uint64"), False),
        ("float16", True),
        (np.float16, True),
        (np.dtype("float16"), True),
        (torch.float16, True),
        ("float32", True),
        (np.float32, True),
        (np.dtype("float32"), True),
        (torch.float32, True),
        ("float64", True),
        (np.float64, True),
        (np.dtype("float64"), True),
        (torch.float64, True),
        ("bool", False),
        (bool, False),
        (np.bool_, False),
        (torch.bool, False),
    ],
)
def test_is_dtype_float(type_, expected):
    assert misc.is_dtype_float(type_) is expected


@pytest.mark.parametrize(
    "type_,expected",
    [
        (Any, False),
        (object, False),
        ("Any", False),
        ("object", False),
        ("O", False),
        ("int8", False),
        (np.int8, False),
        (np.dtype("int8"), False),
        (torch.int8, False),
        ("int16", False),
        (np.int16, False),
        (np.dtype("int16"), False),
        (torch.int16, False),
        ("int32", False),
        (np.int32, False),
        (np.dtype("int32"), False),
        (torch.int32, False),
        ("int64", False),
        (np.int64, False),
        (np.dtype("int64"), False),
        (torch.int64, False),
        ("uint8", False),
        (np.uint8, False),
        (np.dtype("uint8"), False),
        (torch.uint8, False),
        ("uint16", False),
        (np.uint16, False),
        (np.dtype("uint16"), False),
        ("uint32", False),
        (np.uint32, False),
        (np.dtype("uint32"), False),
        ("uint64", False),
        (np.uint64, False),
        (np.dtype("uint64"), False),
        ("float16", False),
        (np.float16, False),
        (np.dtype("float16"), False),
        (torch.float16, False),
        ("float32", False),
        (np.float32, False),
        (np.dtype("float32"), False),
        (torch.float32, False),
        ("float64", False),
        (np.float64, False),
        (np.dtype("float64"), False),
        (torch.float64, False),
        ("bool", True),
        (bool, True),
        (np.bool_, True),
        (torch.bool, True),
    ],
)
def test_is_dtype_bool(type_, expected):
    assert misc.is_dtype_bool(type_) is expected


@pytest.mark.parametrize(
    "type_,expected",
    [
        (Any, False),
        (object, False),
        ("Any", False),
        ("object", False),
        ("O", False),
        ("int8", True),
        (np.int8, True),
        (np.dtype("int8"), True),
        (torch.int8, True),
        ("int16", True),
        (np.int16, True),
        (np.dtype("int16"), True),
        (torch.int16, True),
        ("int32", True),
        (np.int32, True),
        (np.dtype("int32"), True),
        (torch.int32, True),
        ("int64", True),
        (np.int64, True),
        (np.dtype("int64"), True),
        (torch.int64, True),
        ("uint8", True),
        (np.uint8, True),
        (np.dtype("uint8"), True),
        (torch.uint8, True),
        ("uint16", True),
        (np.uint16, True),
        (np.dtype("uint16"), True),
        ("uint32", True),
        (np.uint32, True),
        (np.dtype("uint32"), True),
        ("uint64", True),
        (np.uint64, True),
        (np.dtype("uint64"), True),
        ("float16", True),
        (np.float16, True),
        (np.dtype("float16"), True),
        (torch.float16, True),
        ("float32", True),
        (np.float32, True),
        (np.dtype("float32"), True),
        (torch.float32, True),
        ("float64", True),
        (np.float64, True),
        (np.dtype("float64"), True),
        (torch.float64, True),
        ("bool", False),
        (bool, False),
        (np.bool_, False),
        (torch.bool, False),
    ],
)
def test_is_dtype_real(type_, expected):
    assert misc.is_dtype_real(type_) is expected


@pytest.mark.parametrize(
    "obj,expected",
    [
        (10, True),
        (np.int8(10), True),
        (np.int16(10), True),
        (np.int32(10), True),
        (np.int64(10), True),
        (np.uint8(10), True),
        (np.uint16(10), True),
        (np.uint32(10), True),
        (np.uint64(10), True),
        (np.array(10), True),
        (torch.tensor(10), True),
        (10.0, False),
        (np.float16(10.0), False),
        (np.float32(10.0), False),
        (np.float64(10.0), False),
        (np.array(10.0), False),
        (torch.tensor(10.0), False),
        (True, False),
        (False, False),
        (np.array(False), False),
        (torch.tensor(False), False),
        (np.array([1, 2]), False),
        (torch.tensor([1, 2]), False),
        (np.array([1.0, 2.0]), False),
        (torch.tensor([1.0, 2.0]), False),
    ],
)
def test_is_integer(obj, expected):
    assert misc.is_integer(obj) is expected


@pytest.mark.parametrize(
    "obj,expected",
    [
        (10, True),
        (np.int8(10), True),
        (np.int16(10), True),
        (np.int32(10), True),
        (np.int64(10), True),
        (np.uint8(10), True),
        (np.uint16(10), True),
        (np.uint32(10), True),
        (np.uint64(10), True),
        (np.array(10), True),
        (torch.tensor(10), True),
        (10.0, True),
        (np.float16(10.0), True),
        (np.float32(10.0), True),
        (np.float64(10.0), True),
        (np.array(10.0), True),
        (torch.tensor(10.0), True),
        (True, False),
        (False, False),
        (np.array(False), False),
        (torch.tensor(False), False),
        (np.array([1, 2]), False),
        (torch.tensor([1, 2]), False),
        (np.array([1.0, 2.0]), False),
        (torch.tensor([1.0, 2.0]), False),
    ],
)
def test_is_real(obj, expected):
    assert misc.is_real(obj) is expected


@pytest.mark.parametrize(
    "obj,expected",
    [
        (10, False),
        (np.int8(10), False),
        (np.int16(10), False),
        (np.int32(10), False),
        (np.int64(10), False),
        (np.uint8(10), False),
        (np.uint16(10), False),
        (np.uint32(10), False),
        (np.uint64(10), False),
        (np.array(10), False),
        (torch.tensor(10), False),
        (10.0, False),
        (np.float16(10.0), False),
        (np.float32(10.0), False),
        (np.float64(10.0), False),
        (np.array(10.0), False),
        (torch.tensor(10.0), False),
        (True, True),
        (False, True),
        (np.array(False), True),
        (torch.tensor(False), True),
        (np.array([False, False]), False),
        (torch.tensor([False, False]), False),
        (np.array([1, 2]), False),
        (torch.tensor([1, 2]), False),
        (np.array([1.0, 2.0]), False),
        (torch.tensor([1.0, 2.0]), False),
    ],
)
def test_is_bool(obj, expected):
    assert misc.is_bool(obj) is expected


@pytest.mark.parametrize(
    "obj,expected",
    [
        (1, False),
        (np.array(1), False),
        (torch.tensor(1), False),
        ([1, 2, 3], True),
        ((1, 2, 3), True),
        (np.array([1, 2, 3]), True),
        (torch.tensor([1, 2, 3]), True),
        (np.array([[1, 2], [3, 4]]), False),
        (torch.tensor([[1, 2], [3, 4]]), False),
        (np.array(1.0), False),
        (torch.tensor(1.0), False),
        ([1.0, 2.0, 3.0], False),
        ((1.0, 2.0, 3.0), False),
        (np.array([1.0, 2.0, 3.0]), False),
        (torch.tensor([1.0, 2.0, 3.0]), False),
        (np.array([[1.0, 2.0], [3.0, 4.0]]), False),
        (torch.tensor([[1.0, 2.0], [3.0, 4.0]]), False),
        (np.array(True), False),
        (torch.tensor(True), False),
        ([True, False, True], False),
        ((True, False, True), False),
        (np.array([True, False, True]), False),
        (torch.tensor([True, False, True]), False),
        (np.array([[True, False], [True, False]]), False),
        (torch.tensor([[True, False], [True, False]]), False),
    ],
)
def test_is_integer_vector(obj, expected):
    assert misc.is_integer_vector(obj) is expected


@pytest.mark.parametrize(
    "obj,expected",
    [
        (1, False),
        (np.array(1), False),
        (torch.tensor(1), False),
        ([1, 2, 3], True),
        ((1, 2, 3), True),
        (np.array([1, 2, 3]), True),
        (torch.tensor([1, 2, 3]), True),
        (np.array([[1, 2], [3, 4]]), False),
        (torch.tensor([[1, 2], [3, 4]]), False),
        (np.array(1.0), False),
        (torch.tensor(1.0), False),
        ([1.0, 2.0, 3.0], True),
        ((1.0, 2.0, 3.0), True),
        (np.array([1.0, 2.0, 3.0]), True),
        (torch.tensor([1.0, 2.0, 3.0]), True),
        (np.array([[1.0, 2.0], [3.0, 4.0]]), False),
        (torch.tensor([[1.0, 2.0], [3.0, 4.0]]), False),
        (np.array(True), False),
        (torch.tensor(True), False),
        ([True, False, True], False),
        ((True, False, True), False),
        (np.array([True, False, True]), False),
        (torch.tensor([True, False, True]), False),
        (np.array([[True, False], [True, False]]), False),
        (torch.tensor([[True, False], [True, False]]), False),
    ],
)
def test_is_real_vector(obj, expected):
    assert misc.is_real_vector(obj) is expected


@pytest.mark.parametrize(
    "obj,expected",
    [
        (1, False),
        (np.array(1), False),
        (torch.tensor(1), False),
        ([1, 2, 3], False),
        ((1, 2, 3), False),
        (np.array([1, 2, 3]), False),
        (torch.tensor([1, 2, 3]), False),
        (np.array([[1, 2], [3, 4]]), False),
        (torch.tensor([[1, 2], [3, 4]]), False),
        (np.array(1.0), False),
        (torch.tensor(1.0), False),
        ([1.0, 2.0, 3.0], False),
        ((1.0, 2.0, 3.0), False),
        (np.array([1.0, 2.0, 3.0]), False),
        (torch.tensor([1.0, 2.0, 3.0]), False),
        (np.array([[1.0, 2.0], [3.0, 4.0]]), False),
        (torch.tensor([[1.0, 2.0], [3.0, 4.0]]), False),
        (np.array(True), False),
        (torch.tensor(True), False),
        ([True, False, True], True),
        ((True, False, True), True),
        (np.array([True, False, True]), True),
        (torch.tensor([True, False, True]), True),
        (np.array([[True, False], [True, False]]), False),
        (torch.tensor([[True, False], [True, False]]), False),
    ],
)
def test_is_bool_vector(obj, expected):
    assert misc.is_bool_vector(obj) is expected


def test_ensure_tensor_length_and_dtype():
    inputs = (
        [1, 2, 3],
        (1, 2, 3),
        np.array([1, 2, 3]),
        torch.tensor([1, 2, 3]),
    )

    for x in inputs:
        t = misc.ensure_tensor_length_and_dtype(x, length=3, dtype="float32")
        assert t.dtype is torch.float32
        assert len(t) == 3

    scalar_ones = (1, np.array(1), torch.tensor(1))

    for x in scalar_ones:
        t = misc.ensure_tensor_length_and_dtype(x, length=3, dtype="int64")
        assert t.dtype is torch.int64
        assert len(t) == 3
        for element in t:
            assert int(element) == 1

    for x in scalar_ones:
        t = misc.ensure_tensor_length_and_dtype(x, length=3, allow_scalar=True, dtype="int64")
        assert t.dtype is torch.int64
        assert t.ndim == 0
        assert int(t) == 1

    erroneous_inputs = (
        [1, 2],
        (1, 2),
        [1, 2, 3, 4],
        (1, 2, 3, 4),
        [
            [1, 2, 3],
            [4, 5, 6],
        ],
        torch.tensor(
            [
                [1, 2],
                [3, 4],
            ]
        ),
        np.array(
            [
                [1, 2],
                [3, 4],
            ]
        ),
    )

    for x in erroneous_inputs:
        with pytest.raises(ValueError):
            misc.ensure_tensor_length_and_dtype(x, length=3, dtype="float32")


@pytest.mark.parametrize(
    "x,lb,ub,expected",
    [
        (FloatTensor([1, -1, 3]), 0, 2, FloatTensor([1, 0, 2])),
        (FloatTensor([1, -1, 3]), 0, None, FloatTensor([1, 0, 3])),
        (FloatTensor([1, -1, 3]), None, 2, FloatTensor([1, -1, 2])),
        (
            FloatTensor([1, -1, 3]),
            FloatTensor([1.5, 0, -1]),
            FloatTensor([2.5, 1, 2.5]),
            FloatTensor([1.5, 0, 2.5]),
        ),
    ],
)
def test_clip_tensor(x, lb, ub, expected):
    actual = misc.clip_tensor(x, lb, ub)
    assert_allclose(actual, expected, atol=0.00001)


# fmt: off
@pytest.mark.parametrize(
    "x,target,lb,ub,max_change,expected",
    [
        (
            FloatTensor([10, 11, 12]),   # x
            FloatTensor([0, 21, 22]),    # target
            None,                        # lb
            None,                        # ub
            None,                        # max_change
            FloatTensor([0, 21, 22]),    # expected
        ),
        (
            FloatTensor([10, 11, 12]),   # x
            FloatTensor([0, 21, 22]),    # target
            5,                           # lb
            None,                        # ub
            None,                        # max_change
            FloatTensor([5, 21, 22]),    # expected
        ),
        (
            FloatTensor([10, 11, 12]),   # x
            FloatTensor([0, 21, 22]),    # target
            5,                           # lb
            20,                          # ub
            None,                        # max_change
            FloatTensor([5, 20, 20]),    # expected
        ),
        (
            FloatTensor([10, 11, 12]),   # x
            FloatTensor([0, 21, 22]),    # target
            None,                        # lb
            None,                        # ub
            0.5,                         # max_change
            FloatTensor([5, 16.5, 18]),  # expected
        ),
        (
            FloatTensor([10, 11, 12]),   # x
            FloatTensor([0, 21, 22]),    # target
            7,                           # lb
            17,                          # ub
            0.5,                         # max_change
            FloatTensor([7, 16.5, 17]),  # expected
        ),
    ],
)
def test_modify_tensor(x, target, lb, ub, max_change, expected):
    actual = misc.modify_tensor(x, target, lb=lb, ub=ub, max_change=max_change)
    assert_allclose(actual, expected, atol=0.00001)
# fmt: on


@pytest.mark.parametrize(
    "x,shape,length,dtype",
    [
        (FloatTensor([1, 2, 3, 4]), None, None, None),
        (FloatTensor([1, 2, 3, 4]), (5, 5), None, None),
        (FloatTensor([1, 2, 3, 4]), None, 20, None),
        (FloatTensor([1, 2, 3, 4]), None, None, torch.float16),
        (FloatTensor([1, 2, 3, 4]), None, None, torch.float16),
        (ObjectArray(2), (5,), None, None),
        (ObjectArray(2), (5, 3), None, None),
        (ObjectArray(2), None, 20, None),
    ],
)
def test_empty_tensor_like(x, shape, length, dtype):
    def do():
        return misc.empty_tensor_like(x, shape=shape, length=length, dtype=dtype)

    if ((shape is not None) and (length is not None)) or (
        isinstance(x, ObjectArray) and (shape is not None) and (len(shape) >= 2)
    ):
        with pytest.raises(ValueError):
            do()
    else:
        result = do()

        if shape is None and length is None:
            assert result.shape == x.shape
        elif shape is not None and length is not None:
            assert False, "Should have raised error"
        elif shape is not None:
            assert result.shape == shape
        elif length is not None:
            assert result.ndim == 1
            assert len(result) == length

        if dtype is None:
            assert result.dtype == x.dtype
        else:
            assert result.dtype == dtype


@pytest.mark.parametrize("x", [FloatTensor([1, 2, 3]), [1, 2, 3], (1, 2, 3), np.array([1.0, 2, 3])])
def test_numpy_copy(x):
    for type_ in ("float16", "float32"):
        copied = misc.numpy_copy(x, type_)
        copied[:] = 0

        assert_allclose(x, FloatTensor([1, 2, 3]), atol=0.00001)
        assert_dtype_matches(copied, type_)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for reading and for writing neural network parameters"""

from typing import Optional

import torch
from torch import nn

from ...tools.misc import Device


@torch.no_grad()
def fill_parameters(net: nn.Module, vector: torch.Tensor):
    """Fill the parameters of a torch module (net) from a vector.

    No gradient information is kept.

    The vector's length must be exactly the same with the number
    of parameters of the PyTorch module.

    Args:
        net: The torch module whose parameter values will be filled.
        vector: A 1-D torch tensor which stores the parameter values.
    """
    address = 0
    for p in net.parameters():
        d = p.data.view(-1)
        n = len(d)
        d[:] = torch.as_tensor(vector[address : address + n], device=d.device)
        address += n

    if address != len(vector):
        raise IndexError("The parameter vector is larger than expected")


@torch.no_grad()
def parameter_vector(net: nn.Module, *, device: Optional[Device] = None) -> torch.Tensor:
    """Get all the parameters of a torch module (net) into a vector

    No gradient information is kept.

    Args:
        net: The torch module whose parameters will be extracted.
        device: The device in which the parameter vector will be constructed.
            If the network has parameter across multiple devices,
            you can specify this argument so that concatenation of all the
            parameters will be successful.
    Returns:
        The parameters of the module in a 1-D tensor.
    """
    dev_kwarg = {} if device is None else {"device": device}

    all_vectors = []
    for p in net.parameters():
        all_vectors.append(torch.as_tensor(p.data.view(-1), **dev_kwarg))

    return torch.cat(all_vectors)


def count_parameters(net: nn.Module) -> int:
    """
    Get the number of parameters the network.

    Args:
        net: The torch module whose parameters will be counted.
    Returns:
        The number of parameters, as an integer.
    """

    count = 0

    for p in net.parameters():
        count += p.numel()

    return count



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains ranking functions which work with PyTorch tensors.
"""

from typing import Iterable

import torch


def centered(fitnesses: torch.Tensor, *, higher_is_better: bool = True) -> torch.Tensor:
    """
    Apply linearly spaced 0-centered ranking on a PyTorch tensor.
    The lowest weight is -0.5, and the highest weight is 0.5.
    This is the same ranking method that was used in:

        Tim Salimans, Jonathan Ho, Xi Chen, Szymon Sidor, Ilya Sutskever (2017).
        Evolution Strategies as a Scalable Alternative to Reinforcement Learning

    Args:
        fitnesses: A PyTorch tensor which contains real numbers which we want
             to rank.
        higher_is_better: Whether or not the higher values will be assigned
             higher ranks. Changing this to False means that lower values
             are interpreted as better, and therefore lower values will have
             higher ranks.
    Returns:
        The ranks, in the same device, with the same dtype with the original
        tensor.
    """
    device = fitnesses.device
    dtype = fitnesses.dtype
    with torch.no_grad():
        x = fitnesses.reshape(-1)
        n = len(x)
        indices = x.argsort(descending=(not higher_is_better))
        weights = (torch.arange(n, dtype=dtype, device=device) / (n - 1)) - 0.5
        ranks = torch.empty_like(x)
        ranks[indices] = weights
        return ranks.reshape(*(fitnesses.shape))


def linear(fitnesses: torch.Tensor, *, higher_is_better: bool = True) -> torch.Tensor:
    """
    Apply linearly spaced ranking on a PyTorch tensor.
    The lowest weight is 0, and the highest weight is 1.

    Args:
        fitnesses: A PyTorch tensor which contains real numbers which we want
             to rank.
        higher_is_better: Whether or not the higher values will be assigned
             higher ranks. Changing this to False means that lower values
             are interpreted as better, and therefore lower values will have
             higher ranks.
    Returns:
        The ranks, in the same device, with the same dtype with the original
        tensor.
    """
    device = fitnesses.device
    dtype = fitnesses.dtype
    with torch.no_grad():
        x = fitnesses.reshape(-1)
        n = len(x)
        indices = x.argsort(descending=(not higher_is_better))
        weights = torch.arange(n, dtype=dtype, device=device) / (n - 1)
        ranks = torch.empty_like(x)
        ranks[indices] = weights
        return ranks.reshape(*(fitnesses.shape))


def nes(fitnesses: torch.Tensor, *, higher_is_better: bool = True) -> torch.Tensor:
    """
    Apply the ranking mechanism proposed in:

        Wierstra, D., Schaul, T., Glasmachers, T., Sun, Y., Peters, J., & Schmidhuber, J. (2014).
        Natural evolution strategies. The Journal of Machine Learning Research, 15(1), 949-980.

    Args:
        fitnesses: A PyTorch tensor which contains real numbers which we want
             to rank.
        higher_is_better: Whether or not the higher values will be assigned
             higher ranks. Changing this to False means that lower values
             are interpreted as better, and therefore lower values will have
             higher ranks.
    Returns:
        The ranks, in the same device, with the same dtype with the original
        tensor.
    """
    device = fitnesses.device
    dtype = fitnesses.dtype

    with torch.no_grad():
        x = fitnesses.reshape(-1)
        n = len(x)

        incr_indices = torch.arange(n, dtype=dtype, device=device)
        N = torch.tensor(n, dtype=dtype, device=device)

        weights = torch.max(
            torch.tensor(0, dtype=dtype, device=device), torch.log((N / 2.0) + 1.0) - torch.log(N - incr_indices)
        )

        indices = torch.argsort(x, descending=(not higher_is_better))
        ranks = torch.empty(n, dtype=indices.dtype, device=device)
        ranks[indices] = torch.arange(n, dtype=indices.dtype, device=device)

        utils = weights[ranks]
        utils /= torch.sum(utils)
        utils -= 1 / N

        return utils.reshape(*(fitnesses.shape))


def normalized(fitnesses: torch.Tensor, *, higher_is_better: bool = True) -> torch.Tensor:
    """
    Normalize the fitnesses and return the result as ranks.

    The normalization is done in such a way that the mean becomes 0.0 and
    the standard deviation becomes 1.0.

    According to the value of `higher_is_better`, it will be ensured that
    better solutions will have numerically higher rank.
    In more details, if `higher_is_better` is set as False, then the
    fitnesses will be multiplied by -1.0 in addition to being subject
    to normalization.

    Args:
        fitnesses: A PyTorch tensor which contains real numbers which we want
             to rank.
        higher_is_better: Whether or not the higher values will be assigned
             higher ranks. Changing this to False means that lower values
             are interpreted as better, and therefore lower values will have
             higher ranks.
    Returns:
        The ranks, in the same device, with the same dtype with the original
        tensor.
    """
    if not higher_is_better:
        fitnesses = -fitnesses

    fitness_mean = torch.mean(fitnesses)
    fitness_stdev = torch.std(fitnesses)

    fitnesses = fitnesses - fitness_mean
    fitnesses = fitnesses / fitness_stdev

    return fitnesses


def raw(fitnesses: torch.Tensor, *, higher_is_better: bool = True) -> torch.Tensor:
    """
    Return the fitnesses themselves as ranks.

    If `higher_is_better` is given as False, then the fitnesses will first
    be multiplied by -1 and then the result will be returned as ranks.

    Args:
        fitnesses: A PyTorch tensor which contains real numbers which we want
             to rank.
        higher_is_better: Whether or not the higher values will be assigned
             higher ranks. Changing this to False means that lower values
             are interpreted as better, and therefore lower values will have
             higher ranks.
    Returns:
        The ranks, in the same device, with the same dtype with the original
        tensor.
    """
    if not higher_is_better:
        fitnesses = -fitnesses
    return fitnesses


rankers = {"nes": nes, "centered": centered, "linear": linear, "normalized": normalized, "raw": raw}


def rank(fitnesses: Iterable[float], ranking_method: str, *, higher_is_better: bool):
    """
    Get the ranks of the given sequence of numbers.

    Better solutions will have numerically higher ranks.

    Args:
        fitnesses: A sequence of numbers to be ranked.
        ranking_method: The ranking method to be used.
            Can be "centered", which means 0-centered linear ranking
                from -0.5 to 0.5.
            Can be "linear", which means a linear ranking from 0 to 1.
            Can be "nes", which means the ranking method used by
                Natural Evolution Strategies.
            Can be "normalized", which means that the ranks will be
                the normalized counterparts of the fitnesses.
            Can be "raw", which means that the fitnesses themselves
                (or, if `higher_is_better` is False, their inverted
                counterparts, inversion meaning the operation of
                multiplying by -1 in this context) will be the ranks.
        higher_is_better: Whether or not the higher values will be assigned
             higher ranks. Changing this to False means that lower values
             are interpreted as better, and therefore lower values will have
             higher ranks.
    """
    fitnesses = torch.as_tensor(fitnesses)
    rank_func = rankers[ranking_method]
    return rank_func(fitnesses, higher_is_better=higher_is_better)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
from pathlib import Path

import numpy as np
import ray

SEED = 0


def pytest_sessionstart(session):
    random.seed(SEED)
    np.random.seed(SEED)

    ray.init(
        num_cpus=1,
        log_to_driver=False,
        local_mode=True,
        include_dashboard=False,
        dashboard_port=None,
        object_store_memory=256 * 1024**2,
        _memory=512 * 1024**2,
        _redis_max_memory=256 * 1024**2,
        _system_config={
            "object_timeout_milliseconds": 200,
            "num_heartbeats_timeout": 10,
            "object_store_full_delay_ms": 100,
        },
    )


def pytest_sessionfinish(session, exitstatus):
    ray.shutdown()



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This namespace various RL-specific utilities."""
from typing import Iterable

import gym
import torch
from torch import nn

from .runningstat import RunningStat


def _accumulate(values: Iterable, method: str):
    if method == "sum":
        return sum(values)
    elif method == "min":
        return min(values)
    elif method == "max":
        return max(values)
    elif method == "avg":
        total = None
        count = 0
        for i, value in enumerate(values):
            count += 1
            if i == 0:
                total = value
            else:
                total += value
        return total / count


def _accumulate_across_dicts(dicts: Iterable[dict], key: str, method: str):
    values = []
    for d in dicts:
        values.append(d[key])
    return _accumulate(values, method)


def _accumulate_all_across_dicts(dicts: Iterable[dict], keys_and_methods: dict):
    result = {}
    for key, method in keys_and_methods.items():
        result[key] = _accumulate_across_dicts(dicts, key, method)
    return result


def reset_env(env: gym.Env) -> Iterable:
    """
    Reset a gym environment.

    For gym 1.0, the plan is to have a `reset(...)` method which returns
    a two-element tuple `(observation, info)` where `info` is an object
    providing any additional information regarding the initial state of
    the agent. However, the old (pre 1.0) gym API (and some environments
    which were written with old gym compatibility in mind) has (or have)
    a `reset(...)` method which returns a single object that is the
    initial observation.
    With the assumption that the observation space of the environment
    is NOT tuple, this function can work with both pre-1.0 and (hopefully)
    after-1.0 versions of gym, and always returns the initial observation.

    Please do not use this function on environments whose observation
    spaces or tuples, because then this function cannot distinguish between
    environments whose `reset(...)` methods return a tuple and environments
    whose `reset(...)` methods return a single observation object but that
    observation object is a tuple.

    Args:
        env: The gym environment which will be reset.
    Returns:
        The initial observation
    """
    result = env.reset()
    if isinstance(result, tuple) and (len(result) == 2):
        result = result[0]
    return result


def take_step_in_env(env: gym.Env, action: Iterable) -> tuple:
    """
    Take a step in the gym environment.
    Taking a step means performing the action provided via the arguments.

    For gym 1.0, the plan is to have a `step(...)` method which returns a
    5-elements tuple containing `observation`, `reward`, `terminated`,
    `truncated`, `info` where `terminated` is a boolean indicating whether
    or not the episode is terminated because of the actions taken within the
    environment, and `truncated` is a boolean indicating whether or not the
    episode is finished because the time limit is reached.
    However, the old (pre 1.0) gym API (and some environments which were
    written with old gym compatibility in mind) has (or have) a `step(...)`
    method which returns 4 elements: `observation`, `reward`, `done`, `info`
    where `done` is a boolean indicating whether or not the episode is
    "done", either because of termination or because of truncation.
    This function can work with both pre-1.0 and (hopefully) after-1.0
    versions of gym, and always returns the 4-element tuple as its result.

    Args:
        env: The gym environment in which the given action will be performed.
    Returns:
        A tuple in the form `(observation, reward, done, info)` where
        `observation` is the observation received after performing the action,
        `reward` is the amount of reward gained,
        `done` is a boolean value indicating whether or not the episode has
        ended, and
        `info` is additional information (usually as a dictionary).
    """
    result = env.step(action)
    if isinstance(result, tuple):
        n = len(result)
        if n == 4:
            observation, reward, done, info = result
        elif n == 5:
            observation, reward, terminated, truncated, info = result
            done = terminated or truncated
        else:
            raise ValueError(
                f"The result of the `step(...)` method of the gym environment"
                f" was expected as a tuple of length 4 or 5."
                f" However, the received result is {repr(result)}, which is"
                f" of length {len(result)}."
            )
    else:
        raise TypeError(
            f"The result of the `step(...)` method of the gym environment"
            f" was expected as a tuple of length 4 or 5."
            f" However, the received result is {repr(result)}, which is"
            f" of type {type(result)}."
        )
    return observation, reward, done, info


class ObsNormLayer(nn.Module):
    """Observation normalization layer for a policy network"""

    def __init__(self, stats: RunningStat, trainable_stats: bool):
        """`__init__(...)`: Initialize the observation normalization layer

        Args:
            stats: The RunninStat object storing the mean and stdev of
                all of the observations.
            trainable_stats: Whether or not the normalization data
                are to be stored as trainable parameters.
        """
        nn.Module.__init__(self)

        mean = torch.tensor(stats.mean, dtype=torch.float32)
        stdev = torch.tensor(stats.stdev, dtype=torch.float32)

        if trainable_stats:
            self.obs_mean = nn.Parameter(mean)
            self.obs_stdev = nn.Parameter(stdev)
        else:
            self.obs_mean = mean
            self.obs_stdev = stdev

    def forward(self, x):
        x = x - self.obs_mean
        x = x / self.obs_stdev
        return x


class ActClipLayer(nn.Module):
    def __init__(self, box: gym.spaces.Box):
        nn.Module.__init__(self)

        self.lb = torch.as_tensor(box.low, dtype=torch.float32)
        self.ub = torch.as_tensor(box.high, dtype=torch.float32)

    def forward(self, x):
        return torch.min(torch.max(x, self.lb), self.ub)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Iterable

import pytest
import torch

from evotorch.testing import assert_allclose
from evotorch.tools.ranking import rank


@pytest.mark.parametrize(
    "x",
    [
        [1, 2, 3, 4],
        [4, 3, 2, 1],
        [10, -10, 5, 3],
        [6, 2, 5],
    ],
)
def test_rank(x):
    def do_the_test(x: Iterable, higher_is_better: bool):
        x = torch.tensor(x, dtype=torch.float32)

        if higher_is_better:
            t = x.clone()
        else:
            t = x * -1

        ranks = torch.empty_like(t)
        ranks[t.argsort()] = (torch.arange(len(t)) / (len(t) - 1)) - 0.5

        ranks2 = rank(x, higher_is_better=higher_is_better, ranking_method="centered")

        assert_allclose(ranks, ranks2, atol=0.00001)

    for higher_is_better in (True, False):
        do_the_test(x, higher_is_better)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ast import Import
from collections.abc import Mapping
from typing import Any, Callable, List, Optional, Union
from warnings import warn

import numpy as np
import torch
from torch import nn
from torch.utils.data import DataLoader, Dataset

from ..core import BoundsPairLike, SolutionBatch
from ..tools.misc import Device
from .neproblem import NEProblem


class SupervisedNE(NEProblem):
    """
    Representation of a neuro-evolution problem where the goal is to minimize
    a loss function in a supervised learning setting.

    A supervised learning problem can be defined via subclassing this class
    and overriding the methods
    `_loss(y_hat, y)` (which is to define how the loss is computed)
    and `_make_dataloader()` (which is to define how a new DataLoader is
    created).

    Alternatively, this class can be directly instantiated as follows:

    ```python
    def my_loss_function(output_of_network, desired_output):
        loss = ...  # compute the loss here
        return loss

    problem = SupervisedNE(
        my_dataset,
        MyTorchModuleClass,
        my_loss_function,
        minibatch_size=...,
        ...
    )
    ```
    """

    def __init__(
        self,
        dataset: Dataset,
        network: Union[str, nn.Module, Callable[[], nn.Module]],
        loss_func: Optional[Callable] = None,
        *,
        network_args: Optional[dict] = None,
        initial_bounds: Optional[BoundsPairLike] = (-0.00001, 0.00001),
        minibatch_size: Optional[int] = None,
        num_minibatches: Optional[int] = None,
        num_actors: Optional[Union[int, str]] = "num_devices",
        common_minibatch: bool = True,
        num_gpus_per_actor: Optional[Union[int, float, str]] = None,
        actor_config: Optional[dict] = None,
        num_subbatches: Optional[int] = None,
        subbatch_size: Optional[int] = None,
        device: Optional[Device] = None,
    ):
        """
        `__init__(...)`: Initialize the SupervisedNE.

        Args:
            dataset: The Dataset from which the minibatches will be pulled
            network: A network structure string, or a Callable (which can be
                a class inheriting from `torch.nn.Module`, or a function
                which returns a `torch.nn.Module` instance), or an instance
                of `torch.nn.Module`.
                The object provided here determines the structure of the
                neural network whose parameters will be evolved.
                A network structure string is a string which can be processed
                by `evotorch.neuroevolution.net.str_to_net(...)`.
                Please see the documentation of the function
                `evotorch.neuroevolution.net.str_to_net(...)` to see how such
                a neural network structure string looks like.
            loss_func: Optionally a function (or a Callable object) which
                receives `y_hat` (the output generated by the neural network)
                and `y` (the desired output), and returns the loss as a
                scalar.
                This argument can also be left as None, in which case it will
                be expected that the method `_loss(self, y_hat, y)` is
                overriden by the inheriting class.
            network_args: Optionally a dict-like object, storing keyword
                arguments to be passed to the network while instantiating it.
            initial_bounds: Specifies an interval from which the values of the
                initial neural network parameters will be drawn.
            minibatch_size: Optionally an integer, describing the size of a
                minibatch when pulling data from the dataset.
                Can also be left as None, in which case it will be expected
                that the inheriting class overrides the method
                `_make_dataloader()` and defines how a new DataLoader is to be
                made.
            num_minibatches: An integer, specifying over how many minibatches
                will a single neural network be evaluated.
                If not specified, it will be assumed that the desired number
                of minibatches per network evaluation is 1.
            num_actors: Number of actors to create for parallelized
                evaluation of the solutions.
                Certain string values are also accepted.
                When given as "max" or as "num_cpus", the number of actors
                will be equal to the number of all available CPUs in the ray
                cluster.
                When given as "num_gpus", the number of actors will be
                equal to the number of all available GPUs in the ray
                cluster, and each actor will be assigned a GPU.
                When given as "num_devices", the number of actors will be
                equal to the minimum among the number of CPUs and the number
                of GPUs available in the cluster (or will be equal to the
                number of CPUs if there is no GPU), and each actor will be
                assigned a GPU (if available).
                If `num_actors` is given as "num_gpus" or "num_devices",
                the argument `num_gpus_per_actor` must not be used,
                and the `actor_config` dictionary must not contain the
                key "num_gpus".
                If `num_actors` is given as something other than "num_gpus"
                or "num_devices", and if you wish to assign GPUs to each
                actor, then please see the argument `num_gpus_per_actor`.
            common_minibatch: Whether or not the same minibatches will be
                used when evaluating the solutions.
            actor_config: A dictionary, representing the keyword arguments
                to be passed to the options(...) used when creating the
                ray actor objects. To be used for explicitly allocating
                resources per each actor.
                For example, for declaring that each actor is to use a GPU,
                one can pass `actor_config=dict(num_gpus=1)`.
                Can also be given as None (which is the default),
                if no such options are to be passed.
            num_gpus_per_actor: Number of GPUs to be allocated by each
                remote actor.
                The default behavior is to NOT allocate any GPU at all
                (which is the default behavior of the ray library as well).
                When given as a number `n`, each actor will be given
                `n` GPUs (where `n` can be an integer, or can be a `float`
                for fractional allocation).
                When given as a string "max", then the available GPUs
                across the entire ray cluster (or within the local computer
                in the simplest cases) will be equally distributed among
                the actors.
                When given as a string "all", then each actor will have
                access to all the GPUs (this will be achieved by suppressing
                the environment variable `CUDA_VISIBLE_DEVICES` for each
                actor).
                When the problem is not distributed (i.e. when there are
                no actors), this argument is expected to be left as None.
            num_subbatches: If `num_subbatches` is None (assuming that
                `subbatch_size` is also None), then, when evaluating a
                population, the population will be split into n pieces, `n`
                being the number of actors, and each actor will evaluate
                its assigned piece. If `num_subbatches` is an integer `m`,
                then the population will be split into `m` pieces,
                and actors will continually accept the next unevaluated
                piece as they finish their current tasks.
                The arguments `num_subbatches` and `subbatch_size` cannot
                be given values other than None at the same time.
                While using a distributed algorithm, this argument determines
                how many sub-batches will be generated, and therefore,
                how many gradients will be computed by the remote actors.
            subbatch_size: If `subbatch_size` is None (assuming that
                `num_subbatches` is also None), then, when evaluating a
                population, the population will be split into `n` pieces, `n`
                being the number of actors, and each actor will evaluate its
                assigned piece. If `subbatch_size` is an integer `m`,
                then the population will be split into pieces of size `m`,
                and actors will continually accept the next unevaluated
                piece as they finish their current tasks.
                When there can be significant difference across the solutions
                in terms of computational requirements, specifying a
                `subbatch_size` can be beneficial, because, while one
                actor is busy with a subbatch containing computationally
                challenging solutions, other actors can accept more
                tasks and save time.
                The arguments `num_subbatches` and `subbatch_size` cannot
                be given values other than None at the same time.
                While using a distributed algorithm, this argument determines
                the size of a sub-batch (or sub-population) sampled by a
                remote actor for computing a gradient.
                In distributed mode, it is expected that the population size
                is divisible by `subbatch_size`.
            device: Default device in which a new population will be generated
                and the neural networks will operate.
                If not specified, "cpu" will be used.
        """
        super().__init__(
            objective_sense="min",
            network=network,
            network_args=network_args,
            initial_bounds=initial_bounds,
            num_actors=num_actors,
            num_gpus_per_actor=num_gpus_per_actor,
            actor_config=actor_config,
            num_subbatches=num_subbatches,
            subbatch_size=subbatch_size,
            device=device,
        )

        self.dataset = dataset
        self.dataloader: DataLoader = None

        self._loss_func = loss_func
        self._minibatch_size = None if minibatch_size is None else int(minibatch_size)
        self._num_minibatches = 1 if num_minibatches is None else int(num_minibatches)
        self._common_minibatch = common_minibatch
        self._current_minibatches: Optional[list] = None

    def _make_dataloader(self) -> DataLoader:
        """
        Make a new DataLoader.

        This method, in its default state, does not contain an implementation.
        In the case where the `__init__` of `SupervisedNE` is not provided
        with a minibatch size, it will be expected that this method is
        overriden by the inheriting class and that the operation of creating
        a new DataLoader is defined here.

        Returns:
            The new DataLoader.
        """
        return NotImplementedError

    def make_dataloader(self) -> DataLoader:
        """
        Make a new DataLoader.

        If the `__init__` of `SupervisedNE` was provided with a minibatch size
        via the argument `minibatch_size`, then a new DataLoder will be made
        with that minibatch size.
        Otherwise, it will be expected that the method `_make_dataloader(...)`
        was overriden to contain details regarding how the DataLoader should be
        created, and that method will be executed.

        Returns:
            The created DataLoader.
        """
        if self._minibatch_size is None:
            return self._make_dataloader()
        else:
            return DataLoader(self.dataset, shuffle=True, batch_size=self._minibatch_size)

    def _evaluate_using_minibatch(self, network: nn.Module, batch: Any) -> Union[float, torch.Tensor]:
        """
        Pass a minibatch through a network, and compute the loss.

        Args:
            network: The network using which the loss will be computed.
            batch: The minibatch that will be used as data.
        Returns:
            The loss.
        """
        with torch.no_grad():
            x, y = batch
            yhat = network(x)
            return self.loss(yhat, y)

    def _loss(self, y_hat: Any, y: Any) -> Union[float, torch.Tensor]:
        """
        The loss function.

        This method, in its default state, does not contain an implementation.
        In the case where `__init__` of `SupervisedNE` class was not given
        a loss function via the argument `loss_func`, it will be expected
        that this method is overriden by the inheriting class and that the
        operation of computing the loss is defined here.

        Args:
            y_hat: The output estimated by the network
            y: The desired output
        Returns:
            A scalar, representing the loss
        """
        raise NotImplementedError

    def loss(self, y_hat: Any, y: Any) -> Union[float, torch.Tensor]:
        """
        Run the loss function and return the loss.

        If the `__init__` of `SupervisedNE` class was given a loss
        function via the argument `loss_func`, then that loss function
        will be used. Otherwise, it will be expected that the method
        `_loss(...)` is overriden with a loss definition, and that method
        will be used to compute the loss.
        The computed loss will be returned.

        Args:
            y_hat: The output estimated by the network
            y: The desired output
        Returns:
            A scalar, representing the loss
        """
        if self._loss_func is None:
            return self._loss(y_hat, y)
        else:
            return self._loss_func(y_hat, y)

    def _prepare(self) -> None:
        self.dataloader = self.make_dataloader()

    def get_minibatch(self) -> Any:
        """
        Get the next minibatch from the DataLoader.
        """
        if self.dataloader is None:
            self._prepare()

        try:
            batch = next(self.dataloader_iterator)
            if batch is None:
                self.dataloader_iterator = iter(self.dataloader)
                batch = self.get_minibatch()
            else:
                batch = batch
        except Exception:
            self.dataloader_iterator = iter(self.dataloader)
            batch = self.get_minibatch()

        # Move batch to device of network
        return [var.to(self.network_device) for var in batch]

    def _evaluate_network(self, network: nn.Module) -> torch.Tensor:
        loss = 0.0
        for batch_idx in range(self._num_minibatches):
            if not self._common_minibatch:
                self._current_minibatch = self.get_minibatch()
            else:
                self._current_minibatch = self._current_minibatches[batch_idx]
            loss += self._evaluate_using_minibatch(network, self._current_minibatch) / self._num_minibatches
        return loss

    def _evaluate_batch(self, batch: SolutionBatch):
        if self._common_minibatch:
            # If using a common data batch, generate them now and use them for the entire batch of solutions
            self._current_minibatches = [self.get_minibatch() for _ in range(self._num_minibatches)]
        return super()._evaluate_batch(batch)



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from collections.abc import Iterable, Mapping, Sequence, Set
from numbers import Number
from typing import Any, Union

import numpy as np
import torch


def _is_basic_data(x: Any) -> bool:
    return (x is None) or isinstance(x, (str, Number))


def _numpy_array_stores_objects(x: np.ndarray) -> bool:
    return x.dtype == np.dtype(object)


def _is_numpy_array_immutable(x: np.ndarray) -> bool:
    return (not _numpy_array_stores_objects(x)) and (not x.flags["WRITEABLE"])


def is_immutable_container_or_tensor(x: Any) -> bool:
    from .objectarray import ObjectArray
    from .readonlytensor import ReadOnlyTensor

    return (
        isinstance(x, (ImmutableContainer, ReadOnlyTensor))
        or (isinstance(x, np.ndarray) and _is_numpy_array_immutable(x))
        or (isinstance(x, ObjectArray) and x.is_read_only)
    )


def as_immutable(x: Any) -> Any:
    from .objectarray import ObjectArray
    from .readonlytensor import ReadOnlyTensor

    if is_immutable_container_or_tensor(x) or _is_basic_data(x):
        return x
    elif isinstance(x, torch.Tensor):
        return x.clone().as_subclass(ReadOnlyTensor)
    elif isinstance(x, ObjectArray):
        return x.clone().get_read_only_view()
    elif isinstance(x, np.ndarray):
        if _numpy_array_stores_objects(x):
            result = ObjectArray(len(x))
            for i, element in enumerate(x):
                result[i] = element
            return result.get_read_only_view()
        else:
            result = x.copy()
            result.flags["WRITEABLE"] = False
            return result
    elif isinstance(x, Mapping):
        return ImmutableDict(x)
    elif isinstance(x, tuple):
        return ImmutableTuple(x)
    elif isinstance(x, set):
        return ImmutableSet(x)
    elif isinstance(x, Iterable):
        return ImmutableList(x)
    else:
        raise TypeError(
            f"as_immutable(...) encountered an object of unsupported type."
            f" Encountered object: {repr(x)}."
            f" Type of the encountered object: {repr(type(x))}."
        )


def mutable_copy(x: Any) -> Any:
    from .objectarray import ObjectArray
    from .readonlytensor import ReadOnlyTensor

    if _is_basic_data(x):
        return x
    elif is_immutable_container_or_tensor(x):
        if isinstance(x, ImmutableContainer):
            return x.clone()
        elif isinstance(x, np.ndarray):
            return x.copy()
        else:
            return x.clone()
    else:
        raise TypeError(
            f"mutable_copy(...) encountered an object of unsupported type."
            f" Encountered object: {repr(x)}."
            f" Type of the encountered object: {repr(type(x))}."
        )


class ImmutableContainer:
    def clone(self) -> Iterable:
        raise NotImplementedError


class ImmutableSequence(ImmutableContainer, Sequence):
    def __init__(self, x: Iterable):
        self.__data: list = [as_immutable(item) for item in x]

    def __len__(self) -> int:
        return len(self.__data)

    def __getitem__(self, i):
        return self.__data[i]

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__data})"

    def __add__(self, other: Iterable) -> "ImmutableSequence":
        cls = type(self)
        return cls(list(self) + list(other))

    def __radd__(self, other: Iterable) -> "ImmutableSequence":
        cls = type(self)
        return cls(list(other) + list(self))

    def clone(self) -> Iterable:
        pass

    def __eq__(self, other: Iterable) -> bool:
        self_length = len(self)
        self_last = self_length - 1
        other_count = 0
        for i, other_item in enumerate(other):
            other_count += 1
            if i > self_last:
                return False
            if self[i] != other_item:
                return False
        return other_count == self_length

    def __ne__(self, other: Iterable) -> bool:
        return not self.__eq__(other)


class ImmutableList(ImmutableSequence):
    def clone(self) -> list:
        return [mutable_copy(x) for x in self]


class ImmutableTuple(ImmutableSequence):
    def clone(self) -> tuple:
        return tuple([mutable_copy(x) for x in self])


class ImmutableSet(ImmutableContainer, Set):
    def __init__(self, x: Iterable):
        self.__data: set = set([as_immutable(item) for item in x])

    def __contains__(self, x: Any) -> bool:
        return as_immutable(x) in self.__data

    def __iter__(self):
        for item in self.__data:
            yield item

    def __len__(self) -> int:
        return len(self.__data)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.__data})"

    def clone(self) -> Iterable:
        return set([mutable_copy(x) for x in self])


def _acceptable_key(x: Any) -> Union[int, str]:
    from .misc import is_integer

    if is_integer(x):
        return int(x)
    elif isinstance(x, str):
        return x
    elif isinstance(x, (tuple, ImmutableSequence)):
        return tuple([_acceptable_key(k) for k in x])
    else:
        raise TypeError(
            f"The object to be used as a key within an `ImmutableDict` must"
            f" be of one of these types: `int`, `str`, `tuple` (where the"
            f" tuple can only contain `int`, `str` and/or `tuple` instances)."
            f" The received key candidate does not seem to satisfy this"
            f" constraint."
            f" The unnaccepted key candidate is: {repr(x)}."
            f" The type of the unnaccepted key is: {repr(type(x))}."
        )


class ImmutableDict(ImmutableContainer, Mapping):
    def __init__(self, x: Iterable, **kwargs):
        self.__data: OrderedDict = OrderedDict()
        iterator = x.items() if isinstance(x, Mapping) else x
        for k, v in iterator:
            self.__data[_acceptable_key(k)] = as_immutable(v)
        for k, v in kwargs.items():
            self.__data[str(k)] = as_immutable(v)

    def __getitem__(self, k):
        k = _acceptable_key(k)
        return self.__data[k]

    def __iter__(self):
        for k in self.__data.keys():
            yield k

    def __len__(self) -> int:
        return len(self.__data)

    def __repr__(self) -> str:
        type_name = type(self).__name__
        contents = ", ".join([repr(k) + ": " + repr(v) for (k, v) in self.__data.items()])
        return type_name + "({" + contents + "})"

    def clone(self) -> "ImmutableDict":
        result = {}
        for k, v in self.items():
            k = mutable_copy(k)
            v = mutable_copy(v)
            result[k] = v
        return result



# The following code is adapted from:
# https://github.com/openai/evolution-strategies-starter/blob/master/es_distributed/es.py.
#
# The license of OpenAI's original code is as follows:
#
# The MIT License
#
# Copyright (c) 2016 OpenAI (http://openai.com)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from copy import copy, deepcopy
from typing import Union

import numpy as np


class RunningStat:
    """
    Tool for efficiently computing the mean and stdev of arrays.
    The arrays themselves are not stored separately,
    instead, they are accumulated.
    """

    def __init__(self):
        """
        ``__init__(...)``: Initialize the RunningStat.

        In the beginning, the number of arrays is 0,
        and the sum and the sum of squares are set as NaN.
        """
        # self.sum = np.zeros(shape, dtype='float32')
        # self.sumsq = np.full(shape, eps, dtype='float32')
        # self.count = eps
        self.reset()

    def reset(self):
        """
        Reset the RunningStat to its initial state.
        """
        self._sum = float("nan")
        self._sumsq = float("nan")
        self._count = 0

    def _increment(self, s, ssq, c):
        # self.sum += s
        # self.sumsq += ssq
        # self.count += c
        if self._count == 0:
            self._sum = np.array(s, dtype="float32")
            self._sumsq = np.array(ssq, dtype="float32")
        else:
            self._sum += s
            self._sumsq += ssq
        self._count += c

    @property
    def count(self) -> int:
        """
        Get the number of arrays accumulated.
        """
        return self._count

    @property
    def sum(self) -> np.ndarray:
        """
        Get the sum of all accumulated arrays.
        """
        return self._sum

    @property
    def sum_of_squares(self) -> np.ndarray:
        """
        Get the sum of squares of all accumulated arrays.
        """
        return self._sumsq

    @property
    def mean(self) -> np.ndarray:
        """
        Get the mean of all accumulated arrays.
        """
        return self._sum / self._count

    @property
    def stdev(self) -> np.ndarray:
        """
        Get the standard deviation of all accumulated arrays.
        """
        return np.sqrt(np.maximum(self._sumsq / self._count - np.square(self.mean), 1e-2))

    # def _set_from_init(self, init_mean, init_std, init_count):
    #    init_mean = np.asarray(init_mean, dtype='float32')
    #    init_std = np.asarray(init_std, dtype='float32')
    #    self._sum = init_mean * init_count
    #    self._sumsq = (np.square(init_mean) + np.square(init_std)) * init_count
    #    self._count = init_count

    def update(self, x: Union[np.ndarray, "RunningStat"]):
        """
        Accumulate more data into the RunningStat object.
        If the argument is an array, that array is added
        as one more data element.
        If the argument is another RunningStat instance,
        all the stats accumulated by that RunningStat object
        are added into this RunningStat object.
        """
        if isinstance(x, RunningStat):
            if x.count > 0:
                self._increment(x.sum, x.sum_of_squares, x.count)
        else:
            self._increment(x, np.square(x), 1)

    def normalize(self, x: Union[np.ndarray, list]) -> np.ndarray:
        """
        Normalize the array x according to the accumulated stats.
        """
        x = np.array(x, dtype="float32")
        x -= self.mean
        x /= self.stdev
        return x

    def __copy__(self):
        return deepcopy(self)

    def __get_repr(self):
        return "<RunningStat, count: " + str(self._count) + ">"

    def __str__(self):
        return self.__get_repr()

    def __repr__(self):
        return self.__get_repr()



# Copyright 2022 NNAISENSE SA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import subprocess as sp
from typing import Optional

DOCS_DIR = os.path.abspath(os.path.dirname(__file__))
NOTEBOOKS_DIR = os.path.abspath(os.path.join(DOCS_DIR, "..", "examples", "notebooks"))
TARGET_DIR = os.path.join(DOCS_DIR, "examples", "notebooks")
GITHUB_BASE_ADDRESS = "https://github.com/nnaisense/evotorch/tree/master/examples/notebooks"


ALL_NOTEBOOK_DIRS = [
    (NOTEBOOKS_DIR, GITHUB_BASE_ADDRESS),
    (
        os.path.join(NOTEBOOKS_DIR, "Model_Predictive_Control_with_CEM"),
        GITHUB_BASE_ADDRESS + "/Model_Predictive_Control_with_CEM",
    ),
]


def convert_notebook(fname: str, output_dir: str) -> Optional[str]:
    _, notebook_name = os.path.split(fname)

    if notebook_name.lower().endswith(".ipynb"):
        raw_name = notebook_name[:-6]
    else:
        raise ValueError(f"Invalid notebook name: {notebook_name}")

    md_name = raw_name + ".md"

    return_code = sp.call(["jupyter", "nbconvert", fname, "--to", "markdown", "--output-dir", output_dir])
    if return_code == 0:
        return os.path.join(output_dir, md_name)
    else:
        return None


def add_lines_to_text_file(fname: str, lines: list):
    with open(fname, "a") as f:
        for line in lines:
            print(line, file=f)


def convert_notebooks_in_dir(notebook_dir: str, github_dir: str):
    for fname in os.listdir(notebook_dir):
        if fname.lower().endswith(".ipynb"):
            full_fname = os.path.join(notebook_dir, fname)
            full_md_name = convert_notebook(full_fname, TARGET_DIR)
            if full_md_name is not None:
                github_address = github_dir + "/" + fname
                add_lines_to_text_file(
                    full_md_name,
                    [
                        "",
                        "---",
                        "",
                        f"[See this notebook on GitHub]({github_address}){{ .md-button .md-button-primary }}",
                    ],
                )


def replace_in_file(fname: str, from_str: str, to_str: str):
    with open(fname, "r") as f:
        all_text = f.read()
    all_text = all_text.replace(from_str, to_str)
    with open(fname, "w") as f:
        f.write(all_text)


def fix_mpc_links():
    replace_in_file(
        os.path.join(TARGET_DIR, "reacher_mpc.md"),
        "train_forward_model/reacher_train.ipynb",
        f"{GITHUB_BASE_ADDRESS}/Model_Predictive_Control_with_CEM/train_forward_model/reacher_train.ipynb",
    )


def convert_all_notebooks(*args, **kwargs):
    os.makedirs(TARGET_DIR, exist_ok=True)
    for notebook_dir, github_dir in ALL_NOTEBOOK_DIRS:
        convert_notebooks_in_dir(notebook_dir, github_dir)
    fix_mpc_links()
